@@ -12,6 +12,7 @@
 #include "timed-animation.h"
 #include "gesture.h"
 #include "gesture-drag.h"
+#include "gesture-swipe.h"
 #include "layer-shell-effects.h"
 
 #define _XOPEN_SOURCE 700
@@ -30,7 +31,7 @@
 #include "view.h"
 #include "xcursor.h"
 
-#define PHOC_ANIM_SUGGEST_STATE_CHANGE_COLOR    (PhocColor){0.0f, 0.3f, 0.5f, 0.5f}
+#define PHOC_ANIM_SUGGEST_STATE_CHANGE_COLOR    (PhocColor){0.0f, 0.6f, 1.0f, 0.5f}
 #define PHOC_ANIM_DURATION_SUGGEST_STATE_CHANGE 200
 
 enum {
@@ -544,27 +545,6 @@ phoc_cursor_get_property (GObject    *object,
   }
 }
 
-/**
- * cursor_gestures_handle_event:
- *
- * Let gestures associated with a cursor handle an event.
- */
-static void
-cursor_gestures_handle_event (PhocCursor *cursor, const PhocEvent *event, double lx, double ly)
-{
-  GSList *gestures = phoc_cursor_get_gestures (cursor);
-
-  if (gestures == NULL)
-    return;
-
-  for (GSList *elem = gestures; elem; elem = elem->next) {
-    PhocGesture *gesture = PHOC_GESTURE (elem->data);
-
-    g_assert (PHOC_IS_GESTURE (gesture));
-    phoc_gesture_handle_event (gesture, event, lx, ly);
-  }
-}
-
 /**
  * handle_gestures_for_event_at:
  *
@@ -579,8 +559,17 @@ handle_gestures_for_event_at (PhocCursor   *self,
                               gsize         size)
 {
   g_autoptr (PhocEvent) event = phoc_event_new (type, wlr_event, size);
+  GSList *gestures = phoc_cursor_get_gestures (self);
 
-  cursor_gestures_handle_event (self, event, lx, ly);
+  if (gestures == NULL)
+    return;
+
+  for (GSList *elem = gestures; elem; elem = elem->next) {
+    PhocGesture *gesture = PHOC_GESTURE (elem->data);
+
+    g_assert (PHOC_IS_GESTURE (gesture));
+    phoc_gesture_handle_event (gesture, event, lx, ly);
+  }
 }
 
 
@@ -741,7 +730,7 @@ phoc_passthrough_cursor (PhocCursor *self, uint32_t time)
   struct wl_client *client = NULL;
   struct wlr_surface *surface;
 
-  surface = phoc_desktop_surface_at (desktop, self->cursor->x, self->cursor->y, &sx, &sy, &view);
+  surface = phoc_desktop_wlr_surface_at (desktop, self->cursor->x, self->cursor->y, &sx, &sy, &view);
   if (surface)
     client = wl_resource_get_client (surface->resource);
 
@@ -986,10 +975,44 @@ on_drag_cancel (PhocGesture *gesture, gpointer sequence, PhocCursor *self)
 }
 
 
+static void
+on_swipe (PhocGestureSwipe *swipe_gesture, double vx, double vy, gpointer data)
+{
+  PhocCursor *self = PHOC_CURSOR (data);
+  PhocDesktop *desktop = phoc_server_get_desktop (phoc_server_get_default ());
+  PhocLayerSurface *layer_surface;
+  PhocDraggableLayerSurface *drag_surface;
+  PhocEventSequence *sequence;
+  double lx, ly;
+
+  g_assert (PHOC_IS_GESTURE_SWIPE (swipe_gesture));
+
+  sequence = phoc_gesture_get_last_updated_sequence (PHOC_GESTURE (swipe_gesture));
+  if (!phoc_gesture_get_point (PHOC_GESTURE (swipe_gesture), sequence, &lx, &ly)) {
+    g_warning ("Failed to get event point for %p", sequence);
+    return;
+  }
+
+  layer_surface = phoc_desktop_layer_surface_at (desktop, lx, ly, NULL, NULL);
+  if (!layer_surface)
+    return;
+
+  drag_surface = phoc_desktop_get_draggable_layer_surface (desktop, layer_surface);
+  if (!drag_surface)
+    return;
+
+  if (!phoc_draggable_layer_surface_fling (drag_surface, lx, ly, vx, vy))
+    return;
+
+  send_touch_cancel (self->seat, layer_surface->layer_surface->surface);
+}
+
+
 static void
 phoc_cursor_init (PhocCursor *self)
 {
-  g_autoptr (PhocGesture) gesture = NULL;
+  g_autoptr (PhocGestureDrag) drag_gesture = NULL;
+  g_autoptr (PhocGestureSwipe) swipe_gesture = NULL;
   PhocCursorPrivate *priv = phoc_cursor_get_instance_private (self);
 
   self->cursor = wlr_cursor_create ();
@@ -1002,14 +1025,18 @@ phoc_cursor_init (PhocCursor *self)
   /*
    * Drag gesture starting at the current cursor position
    */
-  gesture = PHOC_GESTURE (phoc_gesture_drag_new ());
-  g_object_connect (gesture,
+  drag_gesture = phoc_gesture_drag_new ();
+  g_object_connect (drag_gesture,
                     "signal::drag-begin", on_drag_begin, self,
                     "signal::drag-update", on_drag_update, self,
                     "signal::drag-end", on_drag_end, self,
                     "signal::cancel", on_drag_cancel, self,
                     NULL);
-  phoc_cursor_add_gesture (self, gesture);
+  phoc_cursor_add_gesture (self, PHOC_GESTURE (drag_gesture));
+
+  swipe_gesture = phoc_gesture_swipe_new ();
+  g_signal_connect (swipe_gesture, "swipe", G_CALLBACK (on_swipe), self);
+  phoc_cursor_add_gesture (self, PHOC_GESTURE (swipe_gesture));
 }
 
 
@@ -1127,7 +1154,7 @@ phoc_cursor_press_button (PhocCursor              *self,
   PhocView *view;
   struct wlr_surface *surface;
 
-  surface = phoc_desktop_surface_at (desktop, lx, ly, &sx, &sy, &view);
+  surface = phoc_desktop_wlr_surface_at (desktop, lx, ly, &sx, &sy, &view);
   if (state == WLR_BUTTON_PRESSED && view && phoc_seat_grab_meta_press (seat)) {
     phoc_seat_set_focus_view (seat, view);
 
@@ -1183,93 +1210,88 @@ phoc_cursor_press_button (PhocCursor              *self,
 }
 
 
+static void
+phoc_cursor_pointer_motion (PhocCursor              *self,
+                            struct wlr_input_device *device,
+                            double                   dx,
+                            double                   dy,
+                            double                   dx_unaccel,
+                            double                   dy_unaccel,
+                            guint32                  time_msec)
+{
+  PhocDesktop *desktop = phoc_server_get_desktop (phoc_server_get_default ());
+
+  phoc_desktop_notify_activity (desktop, self->seat);
+
+  wlr_relative_pointer_manager_v1_send_relative_motion (desktop->relative_pointer_manager,
+                                                        self->seat->seat,
+                                                        (uint64_t)time_msec * 1000,
+                                                        dx, dy,
+                                                        dx_unaccel, dy_unaccel);
+
+  if (self->active_constraint && device->type == WLR_INPUT_DEVICE_POINTER) {
+    struct wlr_surface *wlr_surface;
+    double sx, sy, sx_out, sy_out;
+
+    wlr_surface = phoc_desktop_wlr_surface_at (desktop,
+                                               self->cursor->x, self->cursor->y,
+                                               &sx, &sy,
+                                               NULL);
+
+    if (self->active_constraint->surface != wlr_surface)
+      return;
+
+    if (!wlr_region_confine (&self->confine, sx, sy, sx + dx, sy + dy, &sx_out, &sy_out))
+      return;
+
+    dx = sx_out - sx;
+    dy = sy_out - sy;
+  }
+
+  wlr_cursor_move (self->cursor, device, dx, dy);
+  phoc_cursor_update_position (self, time_msec);
+}
+
+
 static void
 handle_pointer_motion_relative (struct wl_listener *listener, void *data)
 {
-  PhocServer *server = phoc_server_get_default ();
-  PhocDesktop *desktop = phoc_server_get_desktop (server);
   PhocCursor *self = wl_container_of (listener, self, motion);
   struct wlr_pointer_motion_event *event = data;
   double dx = event->delta_x;
   double dy = event->delta_y;
 
-  double dx_unaccel = event->unaccel_dx;
-  double dy_unaccel = event->unaccel_dy;
-
-  phoc_desktop_notify_activity (desktop, self->seat);
-
-  wlr_relative_pointer_manager_v1_send_relative_motion (
-    desktop->relative_pointer_manager,
-    self->seat->seat, (uint64_t)event->time_msec * 1000, dx, dy,
-    dx_unaccel, dy_unaccel);
-
-  if (self->active_constraint) {
-    PhocView *view = self->pointer_view->view;
-    g_assert (view);
-
-    double lx1 = self->cursor->x;
-    double ly1 = self->cursor->y;
-
-    double lx2 = lx1 + dx;
-    double ly2 = ly1 + dy;
-
-    double sx1 = lx1 - view->box.x;
-    double sy1 = ly1 - view->box.y;
-
-    double sx2 = lx2 - view->box.x;
-    double sy2 = ly2 - view->box.y;
-
-    double sx2_confined, sy2_confined;
-    if (!wlr_region_confine (&self->confine, sx1, sy1, sx2, sy2,
-                             &sx2_confined, &sy2_confined)) {
-      return;
-    }
-
-    dx = sx2_confined - sx1;
-    dy = sy2_confined - sy1;
-  }
-
-  wlr_cursor_move (self->cursor, &event->pointer->base, dx, dy);
-  phoc_cursor_update_position (self, event->time_msec);
+  phoc_cursor_pointer_motion (self,
+                              &event->pointer->base,
+                              dx,
+                              dy,
+                              event->unaccel_dx,
+                              event->unaccel_dy,
+                              event->time_msec);
 }
 
 
 static void
 handle_pointer_motion_absolute (struct wl_listener *listener, void *data)
 {
-  PhocServer *server = phoc_server_get_default ();
-  PhocDesktop *desktop = phoc_server_get_desktop (server);
   PhocCursor *self = wl_container_of (listener, self, motion_absolute);
   struct wlr_pointer_motion_absolute_event *event = data;
-  double lx, ly;
+  double dx, dy, lx, ly;
 
-  phoc_desktop_notify_activity (desktop, self->seat);
-  wlr_cursor_absolute_to_layout_coords (self->cursor, &event->pointer->base, event->x,
-                                        event->y, &lx, &ly);
-
-  double dx = lx - self->cursor->x;
-  double dy = ly - self->cursor->y;
+  wlr_cursor_absolute_to_layout_coords (self->cursor,
+                                        &event->pointer->base,
+                                        event->x, event->y,
+                                        &lx, &ly);
 
   handle_gestures_for_event_at (self, lx, ly, PHOC_EVENT_MOTION_NOTIFY, event, sizeof (*event));
 
-  wlr_relative_pointer_manager_v1_send_relative_motion (
-    desktop->relative_pointer_manager,
-    self->seat->seat, (uint64_t)event->time_msec * 1000, dx, dy, dx, dy);
+  dx = lx - self->cursor->x;
+  dy = ly - self->cursor->y;
 
-  if (self->pointer_view) {
-    PhocView *view = self->pointer_view->view;
-
-    if (self->active_constraint &&
-        !pixman_region32_contains_point (&self->confine,
-                                         floor (lx - view->box.x), floor (ly - view->box.y), NULL)) {
-      return;
-    }
-  }
-
-  wlr_cursor_warp_closest (self->cursor, &event->pointer->base, lx, ly);
-  phoc_cursor_update_position (self, event->time_msec);
+  phoc_cursor_pointer_motion (self, &event->pointer->base, dx, dy, dx, dy, event->time_msec);
 }
 
+
 static void
 handle_pointer_button (struct wl_listener *listener, void *data)
 {
@@ -1360,7 +1382,7 @@ phoc_cursor_handle_touch_down (PhocCursor                  *self,
 
   double sx, sy;
   PhocView *view;
-  struct wlr_surface *surface = phoc_desktop_surface_at (desktop, lx, ly, &sx, &sy, &view);
+  struct wlr_surface *surface = phoc_desktop_wlr_surface_at (desktop, lx, ly, &sx, &sy, &view);
   bool shell_revealed = phoc_handle_shell_reveal (surface, lx, ly, PHOC_SHELL_REVEAL_TOUCH_THRESHOLD);
 
   if (!shell_revealed && surface && phoc_seat_allow_input (seat, surface->resource)) {
@@ -1514,7 +1536,7 @@ phoc_cursor_handle_touch_motion (PhocCursor                    *self,
         found = true;
       } else {
         // FIXME: buggy fallback, but at least handles xdg_popups for now...
-        surface = phoc_desktop_surface_at (desktop, lx, ly, &sx, &sy, NULL);
+        surface = phoc_desktop_wlr_surface_at (desktop, lx, ly, &sx, &sy, NULL);
       }
     }
 
@@ -1612,23 +1634,19 @@ void
 phoc_cursor_handle_request_set_cursor (PhocCursor                                       *self,
                                        struct wlr_seat_pointer_request_set_cursor_event *event)
 {
-  struct wlr_surface *focused_surface =
-    event->seat_client->seat->pointer_state.focused_surface;
-  bool has_focused =
-    focused_surface != NULL && focused_surface->resource != NULL;
+  struct wlr_surface *focused_surface = event->seat_client->seat->pointer_state.focused_surface;
+  bool has_focused = focused_surface != NULL && focused_surface->resource != NULL;
   struct wl_client *focused_client = NULL;
 
-  if (has_focused) {
+  if (has_focused)
     focused_client = wl_resource_get_client (focused_surface->resource);
-  }
-  if (event->seat_client->client != focused_client ||
-      self->mode != PHOC_CURSOR_PASSTHROUGH) {
+
+  if (event->seat_client->client != focused_client || self->mode != PHOC_CURSOR_PASSTHROUGH) {
     g_debug ("Denying request to set cursor from unfocused client");
     return;
   }
 
-  wlr_cursor_set_surface (self->cursor, event->surface, event->hotspot_x,
-                          event->hotspot_y);
+  wlr_cursor_set_surface (self->cursor, event->surface, event->hotspot_x, event->hotspot_y);
   self->cursor_client = event->seat_client->client;
 }
 
@@ -1659,10 +1677,10 @@ phoc_cursor_handle_constraint_commit (PhocCursor *self)
   PhocDesktop *desktop = phoc_server_get_desktop (phoc_server_get_default ());
 
   double sx, sy;
-  struct wlr_surface *surface = phoc_desktop_surface_at (desktop,
-                                                         self->cursor->x,
-                                                         self->cursor->y,
-                                                         &sx, &sy, NULL);
+  struct wlr_surface *surface = phoc_desktop_wlr_surface_at (desktop,
+                                                             self->cursor->x,
+                                                             self->cursor->y,
+                                                             &sx, &sy, NULL);
 
   // This should never happen but views move around right when they're
   // created from (0, 0) to their actual coordinates.
@@ -1690,28 +1708,21 @@ phoc_cursor_constrain (PhocCursor *self,
   if (self->active_constraint == constraint)
     return;
 
-  g_debug ("phoc_cursor_constrain(%p, %p)",
-           self, constraint);
-  g_debug ("self->active_constraint: %p",
-           self->active_constraint);
+  g_debug ("cursor constrain: %p, new: %p, old: %p", self, constraint, self->active_constraint);
 
   wl_list_remove (&self->constraint_commit.link);
   wl_list_init (&self->constraint_commit.link);
-  if (self->active_constraint) {
-    wlr_pointer_constraint_v1_send_deactivated (
-      self->active_constraint);
-  }
+
+  if (self->active_constraint)
+    wlr_pointer_constraint_v1_send_deactivated (self->active_constraint);
 
   self->active_constraint = constraint;
-
   if (constraint == NULL)
     return;
 
   wlr_pointer_constraint_v1_send_activated (constraint);
 
-  wl_list_remove (&self->constraint_commit.link);
-  wl_signal_add (&constraint->surface->events.commit,
-                 &self->constraint_commit);
+  wl_signal_add (&constraint->surface->events.commit, &self->constraint_commit);
   self->constraint_commit.notify = handle_constraint_commit;
 
   pixman_region32_clear (&self->confine);
@@ -1719,23 +1730,24 @@ phoc_cursor_constrain (PhocCursor *self,
   pixman_region32_t *region = &constraint->region;
 
   if (!pixman_region32_contains_point (region, floor (sx), floor (sy), NULL)) {
-    // Warp into region if possible
+    /* Warp into region if possible */
     int nboxes;
     pixman_box32_t *boxes = pixman_region32_rectangles (region, &nboxes);
+
     if (nboxes > 0) {
       PhocView *view = self->pointer_view->view;
 
-      double lx = view->box.x + (boxes[0].x1 + boxes[0].x2) / 2.;
-      double ly = view->box.y + (boxes[0].y1 + boxes[0].y2) / 2.;
+      double lx = view->box.x + (boxes[0].x1 + boxes[0].x2) / 2.0;
+      double ly = view->box.y + (boxes[0].y1 + boxes[0].y2) / 2.0;
 
       wlr_cursor_warp_closest (self->cursor, NULL, lx, ly);
     }
   }
 
-  // A locked pointer will result in an empty region, thus disallowing all movement
-  if (constraint->type == WLR_POINTER_CONSTRAINT_V1_CONFINED) {
+  /* A locked pointer will result in an empty region, thus disallowing
+   * all movement */
+  if (constraint->type == WLR_POINTER_CONSTRAINT_V1_CONFINED)
     pixman_region32_copy (&self->confine, region);
-  }
 }
 
 
