@@ -31,9 +31,9 @@
 #include <wlr/types/wlr_xdg_foreign_v1.h>
 #include <wlr/types/wlr_xdg_foreign_v2.h>
 #include <wlr/types/wlr_xdg_output_v1.h>
-#include <wlr/types/wlr_xdg_output_v1.h>
 #include <wlr/types/wlr_xdg_shell.h>
 #include <wlr/util/box.h>
+
 #include "cursor.h"
 #include "device-state.h"
 #include "idle-inhibit.h"
@@ -41,6 +41,8 @@
 #include "output.h"
 #include "seat.h"
 #include "server.h"
+#include "color-rect.h"
+#include "timed-animation.h"
 #include "utils.h"
 #include "view.h"
 #include "virtual.h"
@@ -49,7 +51,6 @@
 #include "wlr-layer-shell-unstable-v1-protocol.h"
 #include "gesture-swipe.h"
 #include "layer-shell-effects.h"
-
 #include "xdg-surface.h"
 #include "xdg-toplevel-decoration.h"
 #include "xwayland-surface.h"
@@ -59,6 +60,11 @@
 #define PHOC_XDG_SHELL_VERSION 5
 #define PHOC_LAYER_SHELL_VERSION 3
 
+#define PHOC_ANIM_ALWAYS_ON_TOP_DURATION  300
+#define PHOC_ANIM_ALWAYS_ON_TOP_COLOR_ON  (PhocColor){0.5f, 0.0f, 0.3f, 0.5f}
+#define PHOC_ANIM_ALWAYS_ON_TOP_COLOR_OFF (PhocColor){0.3f, 0.5f, 0.3f, 0.5f}
+#define PHOC_ANIM_ALWAYS_ON_TOP_WIDTH     10
+
 /**
  * PhocDesktop:
  *
@@ -198,45 +204,19 @@ layer_surface_at (PhocOutput                     *output,
                   double                         *sx,
                   double                         *sy)
 {
-  PhocLayerSurface *layer_surface;
+  GQueue *layer_surfaces = phoc_output_get_layer_surfaces_for_layer (output, layer);
+
+  for (GList *l = layer_surfaces->tail; l; l = l->prev) {
+    PhocLayerSurface *layer_surface = PHOC_LAYER_SURFACE (l->data);
+    struct wlr_surface *sub;
 
-  /* TODO: use phoc_output_get_layer_surfaces_for_layer */
-  wl_list_for_each_reverse(layer_surface, &output->layer_surfaces, link) {
     if (!layer_surface->mapped)
       continue;
 
-    if (layer_surface->layer != layer)
-      continue;
-
-    if (layer_surface->layer_surface->current.exclusive_zone <= 0)
-      continue;
-
     double _sx = ox - layer_surface->geo.x;
     double _sy = oy - layer_surface->geo.y;
 
-    struct wlr_surface *sub = wlr_layer_surface_v1_surface_at(
-      layer_surface->layer_surface, _sx, _sy, sx, sy);
-
-    if (sub)
-      return sub;
-  }
-
-  wl_list_for_each(layer_surface, &output->layer_surfaces, link) {
-    if (!layer_surface->mapped)
-      continue;
-
-    if (layer_surface->layer != layer)
-      continue;
-
-    if (layer_surface->layer_surface->current.exclusive_zone > 0)
-      continue;
-
-    double _sx = ox - layer_surface->geo.x;
-    double _sy = oy - layer_surface->geo.y;
-
-    struct wlr_surface *sub = wlr_layer_surface_v1_surface_at(
-      layer_surface->layer_surface, _sx, _sy, sx, sy);
-
+    sub = wlr_layer_surface_v1_surface_at (layer_surface->layer_surface, _sx, _sy, sx, sy);
     if (sub)
       return sub;
   }
@@ -245,7 +225,7 @@ layer_surface_at (PhocOutput                     *output,
 }
 
 /**
- * phoc_desktop_surface_at:
+ * phoc_desktop_wlr_surface_at:
  * @desktop: The `PhocDesktop` to look the surface up for
  * @lx: X coordinate the surface to look up at in layout coordinates
  * @ly: Y coordinate the surface to look up at in layout coordinates
@@ -260,68 +240,68 @@ layer_surface_at (PhocOutput                     *output,
  * Returns: (nullable): The `struct wlr_surface`
  */
 struct wlr_surface *
-phoc_desktop_surface_at(PhocDesktop *desktop,
-                        double lx, double ly, double *sx, double *sy,
-                        PhocView **view)
+phoc_desktop_wlr_surface_at (PhocDesktop *desktop,
+                             double       lx,
+                             double       ly,
+                             double      *sx,
+                             double      *sy,
+                             PhocView   **view)
 {
   struct wlr_surface *surface = NULL;
   PhocOutput *output = phoc_desktop_layout_get_output (desktop, lx, ly);
   double ox = lx, oy = ly;
-  if (view) {
+  if (view)
     *view = NULL;
-  }
 
+  if (output)
+    wlr_output_layout_output_coords (desktop->layout, output->wlr_output, &ox, &oy);
+
+  /* Layers above regular views */
   if (output) {
-    wlr_output_layout_output_coords(desktop->layout, output->wlr_output, &ox, &oy);
-
-    if ((surface = layer_surface_at(output, ZWLR_LAYER_SHELL_V1_LAYER_OVERLAY,
-                                    ox, oy, sx, sy))) {
+    surface = layer_surface_at (output, ZWLR_LAYER_SHELL_V1_LAYER_OVERLAY, ox, oy, sx, sy);
+    if (surface)
       return surface;
-    }
-
-    if (output->fullscreen_view != NULL) {
 
+    if (output->fullscreen_view) {
       if (phoc_output_has_shell_revealed (output)) {
-        if ((surface = layer_surface_at(output, ZWLR_LAYER_SHELL_V1_LAYER_TOP,
-                                        ox, oy, sx, sy))) {
+        surface = layer_surface_at (output, ZWLR_LAYER_SHELL_V1_LAYER_TOP, ox, oy, sx, sy);
+        if (surface)
           return surface;
-        }
       }
 
-      if (view_at(output->fullscreen_view, lx, ly, &surface, sx, sy)) {
-        if (view) {
+      if (view_at (output->fullscreen_view, lx, ly, &surface, sx, sy)) {
+        if (view)
           *view = output->fullscreen_view;
-        }
         return surface;
       } else {
         return NULL;
       }
     }
 
-    if ((surface = layer_surface_at(output, ZWLR_LAYER_SHELL_V1_LAYER_TOP,
-                                    ox, oy, sx, sy))) {
+    surface = layer_surface_at (output, ZWLR_LAYER_SHELL_V1_LAYER_TOP, ox, oy, sx, sy);
+    if (surface)
       return surface;
-    }
   }
 
-  PhocView *_view;
-  if ((_view = desktop_view_at(desktop, lx, ly, &surface, sx, sy))) {
-    if (view) {
+  PhocView *_view = desktop_view_at (desktop, lx, ly, &surface, sx, sy);
+  if (_view) {
+    if (view)
       *view = _view;
-    }
+
     return surface;
   }
 
+  /* Layers below regular views */
   if (output) {
-    if ((surface = layer_surface_at(output, ZWLR_LAYER_SHELL_V1_LAYER_BOTTOM,
-                                    ox, oy, sx, sy))) {
+    surface = layer_surface_at (output, ZWLR_LAYER_SHELL_V1_LAYER_BOTTOM, ox, oy, sx, sy);
+    if (surface)
       return surface;
-    }
-    if ((surface = layer_surface_at(output, ZWLR_LAYER_SHELL_V1_LAYER_BACKGROUND,
-                                    ox, oy, sx, sy))) {
+
+    surface = layer_surface_at (output, ZWLR_LAYER_SHELL_V1_LAYER_BACKGROUND, ox, oy, sx, sy);
+    if (surface)
       return surface;
-    }
   }
+
   return NULL;
 }
 
@@ -445,55 +425,60 @@ input_inhibit_deactivate (struct wl_listener *listener, void *data)
 static void
 handle_constraint_destroy (struct wl_listener *listener, void *data)
 {
-  PhocPointerConstraint *constraint = wl_container_of(listener, constraint, destroy);
+  PhocPointerConstraint *constraint = wl_container_of (listener, constraint, destroy);
   struct wlr_pointer_constraint_v1 *wlr_constraint = data;
-  PhocSeat *seat = wlr_constraint->seat->data;
+  PhocSeat *seat = PHOC_SEAT (wlr_constraint->seat->data);
   PhocCursor *cursor = phoc_seat_get_cursor (seat);
 
-  wl_list_remove(&constraint->destroy.link);
+  wl_list_remove (&constraint->destroy.link);
 
   if (cursor->active_constraint == wlr_constraint) {
-    wl_list_remove(&cursor->constraint_commit.link);
-    wl_list_init(&cursor->constraint_commit.link);
+    wl_list_remove (&cursor->constraint_commit.link);
+    wl_list_init (&cursor->constraint_commit.link);
     cursor->active_constraint = NULL;
 
-    if (wlr_constraint->current.committed &
-        WLR_POINTER_CONSTRAINT_V1_STATE_CURSOR_HINT &&
+    if (wlr_constraint->current.committed & WLR_POINTER_CONSTRAINT_V1_STATE_CURSOR_HINT &&
         cursor->pointer_view) {
       PhocView *view = cursor->pointer_view->view;
       double lx = view->box.x + wlr_constraint->current.cursor_hint.x;
       double ly = view->box.y + wlr_constraint->current.cursor_hint.y;
 
-      wlr_cursor_warp(cursor->cursor, NULL, lx, ly);
+      wlr_cursor_warp (cursor->cursor, NULL, lx, ly);
     }
   }
 
-  free(constraint);
+  g_free (constraint);
 }
 
+
 static void
 handle_pointer_constraint (struct wl_listener *listener, void *data)
 {
   PhocDesktop *desktop = phoc_server_get_desktop (phoc_server_get_default ());
   struct wlr_pointer_constraint_v1 *wlr_constraint = data;
-  PhocSeat *seat = wlr_constraint->seat->data;
+  PhocSeat *seat = PHOC_SEAT (wlr_constraint->seat->data);
   PhocCursor *cursor = phoc_seat_get_cursor (seat);
-
-  PhocPointerConstraint *constraint = g_new0 (PhocPointerConstraint, 1);
-  constraint->destroy.notify = handle_constraint_destroy;
-  wl_signal_add(&wlr_constraint->events.destroy, &constraint->destroy);
-
+  PhocPointerConstraint *constraint;
+  struct wlr_surface *surface;
   double sx, sy;
-  struct wlr_surface *surface = phoc_desktop_surface_at(
-    desktop,
-    cursor->cursor->x, cursor->cursor->y, &sx, &sy, NULL);
+
+  constraint = g_new0 (PhocPointerConstraint, 1);
+  constraint->destroy.notify = handle_constraint_destroy;
+  wl_signal_add (&wlr_constraint->events.destroy, &constraint->destroy);
+
+  surface = phoc_desktop_wlr_surface_at (desktop,
+                                         cursor->cursor->x,
+                                         cursor->cursor->y,
+                                         &sx, &sy,
+                                         NULL);
 
   if (surface == wlr_constraint->surface) {
     g_assert (!cursor->active_constraint);
-    phoc_cursor_constrain(cursor, wlr_constraint, sx, sy);
+    phoc_cursor_constrain (cursor, wlr_constraint, sx, sy);
   }
 }
 
+
 static void
 auto_maximize_changed_cb (PhocDesktop *self,
                           const gchar *key,
@@ -616,6 +601,8 @@ on_output_destroyed (PhocDesktop *self, PhocOutput *destroyed_output)
   g_assert (PHOC_IS_DESKTOP (self));
   g_assert (PHOC_IS_OUTPUT (destroyed_output));
 
+  wlr_output_layout_remove (self->layout, phoc_output_get_wlr_output (destroyed_output));
+
   g_hash_table_iter_init (&iter, self->input_output_map);
   while (g_hash_table_iter_next (&iter, (gpointer) &input_name,
                                  (gpointer) &output)) {
@@ -814,7 +801,7 @@ phoc_desktop_constructed (GObject *object)
 
   priv->data_control_manager_v1 = wlr_data_control_manager_v1_create (wl_display);
 
-  /* sm.puri.phosh settings */
+  /* sm.puri.phoc settings */
   priv->settings = g_settings_new ("sm.puri.phoc");
   g_signal_connect_swapped (priv->settings, "changed::auto-maximize",
                             G_CALLBACK (auto_maximize_changed_cb), self);
@@ -1108,7 +1095,7 @@ phoc_desktop_layer_surface_at (PhocDesktop *self, double lx, double ly, double *
 
   g_assert (PHOC_IS_DESKTOP (self));
 
-  wlr_surface = phoc_desktop_surface_at (self, lx, ly, &sx_, &sy_, NULL);
+  wlr_surface = phoc_desktop_wlr_surface_at (self, lx, ly, &sx_, &sy_, NULL);
 
   if (!wlr_surface)
     return NULL;
@@ -1199,6 +1186,25 @@ phoc_desktop_get_draggable_layer_surface (PhocDesktop *self, PhocLayerSurface *l
     priv->layer_shell_effects, layer_surface);
 }
 
+/**
+ * phoc_desktop_get_layer_surface_stacks:
+ * @self: The desktop
+ *
+ * Get the list of currently known stacks
+ *
+ * Returns:(transfer none)(element-type PhocStackedLayerSurface): The layer surface stacks
+ */
+GSList *
+phoc_desktop_get_layer_surface_stacks (PhocDesktop *self)
+{
+  PhocDesktopPrivate *priv;
+
+  g_assert (PHOC_IS_DESKTOP (self));
+
+  priv = phoc_desktop_get_instance_private (self);
+  return phoc_layer_shell_effects_get_layer_surface_stacks (priv->layer_shell_effects);
+}
+
 /**
  * phoc_desktop_get_gtk_shell:
  * @self: The `PhocDesktop`
@@ -1419,6 +1425,22 @@ phoc_desktop_remove_view (PhocDesktop *self, PhocView *view)
   return g_queue_remove (priv->views, view);
 }
 
+
+static void
+on_always_on_top_animation_done (PhocTimedAnimation *anim, PhocColorRect *rect)
+{
+  PhocView *view;
+
+  g_assert (PHOC_IS_TIMED_ANIMATION (anim));
+  g_assert (PHOC_IS_COLOR_RECT (rect));
+
+  view = g_object_get_data (G_OBJECT (rect), "view");
+  g_assert (PHOC_IS_VIEW (view));
+
+  phoc_view_remove_bling (view, PHOC_BLING (rect));
+}
+
+
 /**
  * phoc_desktop_set_view_always_on_top:
  * @self: The desktop singleton
@@ -1437,6 +1459,47 @@ phoc_desktop_set_view_always_on_top (PhocDesktop *self, PhocView *view, gboolean
   g_assert (PHOC_IS_DESKTOP (self));
   g_assert (PHOC_IS_VIEW (view));
 
+  if (phoc_desktop_get_enable_animations (self)) {
+    g_autoptr (PhocTimedAnimation) fade_anim = NULL;
+    g_autoptr (PhocPropertyEaser) easer = NULL;
+    g_autoptr (PhocColorRect) rect = NULL;
+    PhocColor color;
+    struct wlr_box rect_box, geom_box;
+
+    color = on_top ? PHOC_ANIM_ALWAYS_ON_TOP_COLOR_ON : PHOC_ANIM_ALWAYS_ON_TOP_COLOR_OFF;
+
+    /* Grow the rect around the view a bit */
+    phoc_view_get_box (view, &rect_box);
+    phoc_view_get_geometry (view, &geom_box);
+    rect_box.x -= PHOC_ANIM_ALWAYS_ON_TOP_WIDTH - geom_box.x;
+    rect_box.y -= PHOC_ANIM_ALWAYS_ON_TOP_WIDTH - geom_box.y;
+    rect_box.width += 2 * PHOC_ANIM_ALWAYS_ON_TOP_WIDTH;
+    rect_box.height += 2 * PHOC_ANIM_ALWAYS_ON_TOP_WIDTH;
+    rect = phoc_color_rect_new ((PhocBox *)&rect_box, &color);
+
+    /* Make sure we end up in the render tree */
+    phoc_view_add_bling (view, PHOC_BLING (rect));
+
+    easer = g_object_new (PHOC_TYPE_PROPERTY_EASER,
+                          "target", rect,
+                          "easing", PHOC_EASING_EASE_OUT_QUAD,
+                          NULL);
+    phoc_property_easer_set_props (easer, "alpha", 1.0, 0.0, NULL);
+    fade_anim = g_object_new (PHOC_TYPE_TIMED_ANIMATION,
+                              "animatable", phoc_view_get_output (view),
+                              "duration", PHOC_ANIM_ALWAYS_ON_TOP_DURATION,
+                              "property-easer", easer,
+                              "dispose-on-done", TRUE,
+                              NULL);
+    phoc_bling_map (PHOC_BLING (rect));
+    g_object_set_data (G_OBJECT (rect), "view", view);
+    g_signal_connect (fade_anim,
+                      "done",
+                      G_CALLBACK (on_always_on_top_animation_done),
+                      rect);
+    phoc_timed_animation_play (fade_anim);
+  }
+
   phoc_view_set_always_on_top (view, on_top);
   phoc_desktop_move_view_to_top (self, view);
 
