@@ -49,16 +49,16 @@ enum {
 static guint signals[N_SIGNALS] = { 0 };
 
 typedef struct _PhocOutputPrivate {
-  PhocOutputShield *shield;
-  PhocRenderer     *renderer;
+  PhocRenderer            *renderer;
+  PhocOutputShield        *shield;
 
-  GSList *frame_callbacks;
-  gint    frame_callback_next_id;
-  gint64  last_frame_us;
+  GSList                  *frame_callbacks;
+  gint                     frame_callback_next_id;
+  gint64                   last_frame_us;
 
-  PhocCutoutsOverlay *cutouts;
-  gulong              render_cutouts_id;
-  struct wlr_texture *cutouts_texture;
+  PhocCutoutsOverlay      *cutouts;
+  gulong                   render_cutouts_id;
+  struct wlr_texture      *cutouts_texture;
 
   gboolean shell_revealed;
   gboolean force_shell_reveal;
@@ -70,6 +70,8 @@ typedef struct _PhocOutputPrivate {
 
   PhocOutputScaleFilter  scale_filter;
   gboolean               gamma_lut_changed;
+
+  GQueue                *layer_surfaces[ZWLR_LAYER_SHELL_V1_LAYER_OVERLAY + 1];
 } PhocOutputPrivate;
 
 static void phoc_output_initable_iface_init (GInitableIface *iface);
@@ -84,6 +86,11 @@ G_DEFINE_TYPE_WITH_CODE (PhocOutput, phoc_output, G_TYPE_OBJECT,
 
 #define PHOC_OUTPUT_SELF(p) PHOC_PRIV_CONTAINER(PHOC_OUTPUT, PhocOutput, (p))
 
+static void phoc_output_for_each_surface (PhocOutput          *self,
+                                          PhocSurfaceIterator  iterator,
+                                          void                *user_data,
+                                          gboolean             visible_only);
+
 typedef struct {
   PhocAnimatable    *animatable;
   PhocFrameCallback  callback;
@@ -260,13 +267,10 @@ update_output_manager_config (PhocDesktop *desktop)
   PhocOutput *output;
 
   wl_list_for_each (output, &desktop->outputs, link) {
-    struct wlr_output_configuration_head_v1 *config_head =
-      wlr_output_configuration_head_v1_create (config, output->wlr_output);
+    struct wlr_output_configuration_head_v1 *config_head;
     struct wlr_box output_box;
 
-    config_head->state.enabled = output->wlr_output->enabled;
-    config_head->state.mode = output->wlr_output->current_mode;
-
+    config_head = wlr_output_configuration_head_v1_create (config, output->wlr_output);
     wlr_output_layout_get_box (output->desktop->layout, output->wlr_output, &output_box);
     if (!wlr_box_empty (&output_box)) {
       config_head->state.x = output_box.x;
@@ -285,8 +289,6 @@ phoc_output_handle_destroy (struct wl_listener *listener, void *data)
   if (self->fullscreen_view)
     phoc_view_set_fullscreen (self->fullscreen_view, false, NULL);
 
-  update_output_manager_config (self->desktop);
-
   g_signal_emit (self, signals[OUTPUT_DESTROY], 0);
 }
 
@@ -765,9 +767,13 @@ phoc_output_fill_state (PhocOutput              *self,
   }
 
   if (output_config && enable) {
+    enum wl_output_transform transform = output_config->transform;
     double scale;
 
     if (wlr_output_is_drm (self->wlr_output)) {
+      if (output_config->drm_panel_orientation)
+        transform = wlr_drm_connector_get_panel_orientation (self->wlr_output);
+
       for (GSList *l = output_config->modes; l; l = l->next) {
         PhocOutputModeConfig *mode_config = l->data;
         wlr_drm_connector_add_mode (self->wlr_output, &mode_config->info);
@@ -776,6 +782,12 @@ phoc_output_fill_state (PhocOutput              *self,
       g_warning ("Can only add modes for DRM backend");
     }
 
+    if (output_config->phys_width)
+      self->wlr_output->phys_width = output_config->phys_width;
+
+    if (output_config->phys_width)
+      self->wlr_output->phys_height = output_config->phys_height;
+
     if (output_config->mode.width)
       phoc_output_state_set_mode (self, pending, output_config);
     else if (preferred_mode != NULL)
@@ -788,9 +800,11 @@ phoc_output_fill_state (PhocOutput              *self,
 
     wlr_output_state_set_scale (pending, adjust_frac_scale (scale));
 
-    wlr_output_state_set_transform (pending, output_config->transform);
+    wlr_output_state_set_transform (pending, transform);
     priv->scale_filter = output_config->scale_filter;
   } else if (enable) {
+    enum wl_output_transform transform = WL_OUTPUT_TRANSFORM_NORMAL;
+
     if (preferred_mode != NULL) {
       g_debug ("Using preferred mode for %s", self->wlr_output->name);
       wlr_output_state_set_mode (pending, preferred_mode);
@@ -809,7 +823,12 @@ phoc_output_fill_state (PhocOutput              *self,
           break;
       }
     }
+
+    if (wlr_output_is_drm (self->wlr_output))
+      transform = wlr_drm_connector_get_panel_orientation (self->wlr_output);
+
     wlr_output_state_set_scale (pending, phoc_output_compute_scale (self, pending));
+    wlr_output_state_set_transform (pending, transform);
   }
 
   if (output_config && output_config->x > 0 && output_config->y > 0) {
@@ -836,15 +855,6 @@ phoc_output_initable_init (GInitable    *initable,
 
   PhocConfig *config = phoc_server_get_config (phoc_server_get_default ());
 
-  g_message ("Output '%s' added ('%s'/'%s'/'%s'), "
-             "%" PRId32 "mm x %" PRId32 "mm",
-             self->wlr_output->name,
-             self->wlr_output->make,
-             self->wlr_output->model,
-             self->wlr_output->serial,
-             self->wlr_output->phys_width,
-             self->wlr_output->phys_height);
-
   self->wlr_output->data = self;
   wl_list_insert (&self->desktop->outputs, &self->link);
 
@@ -908,15 +918,25 @@ phoc_output_initable_init (GInitable    *initable,
     if (priv->cutouts) {
       g_message ("Adding cutouts overlay");
       priv->cutouts_texture = phoc_cutouts_overlay_get_cutouts_texture (priv->cutouts, self);
-      priv->render_cutouts_id =  g_signal_connect_swapped (renderer, "render-end",
-                                                           G_CALLBACK (render_cutouts),
-                                                           self);
+      priv->render_cutouts_id = g_signal_connect_swapped (renderer, "render-end",
+                                                          G_CALLBACK (render_cutouts),
+                                                          self);
     } else {
       g_warning ("Could not create cutout overlay");
     }
   }
 
   wlr_output_state_finish (&pending);
+
+  g_message ("Output '%s' added ('%s'/'%s'/'%s'), "
+             "%" PRId32 "mm x %" PRId32 "mm",
+             self->wlr_output->name,
+             self->wlr_output->make,
+             self->wlr_output->model,
+             self->wlr_output->serial,
+             self->wlr_output->phys_width,
+             self->wlr_output->phys_height);
+
   return TRUE;
 }
 
@@ -926,8 +946,13 @@ phoc_output_finalize (GObject *object)
   PhocOutput *self = PHOC_OUTPUT (object);
   PhocOutputPrivate *priv = phoc_output_get_instance_private (self);
 
+  self->wlr_output->data = NULL;
+  self->wlr_output = NULL;
+
   wl_list_remove (&self->link);
 
+  update_output_manager_config (self->desktop);
+
   wl_list_remove (&self->commit.link);
   wl_list_remove (&self->output_destroy.link);
 
@@ -943,11 +968,13 @@ phoc_output_finalize (GObject *object)
                  (GDestroyNotify)phoc_output_frame_callback_info_free);
 
   wl_list_init (&self->layer_surfaces);
+  for (int i = 0; i < G_N_ELEMENTS (priv->layer_surfaces); i++)
+    g_clear_pointer (&priv->layer_surfaces[i], g_queue_free);
 
+  g_clear_signal_handler (&priv->render_cutouts_id, priv->renderer);
   g_clear_object (&priv->renderer);
   g_clear_object (&priv->cutouts);
   g_clear_pointer (&priv->cutouts_texture, wlr_texture_destroy);
-  g_clear_signal_handler (&priv->render_cutouts_id, self);
   g_clear_object (&priv->shield);
   g_clear_object (&self->desktop);
 
@@ -1206,21 +1233,12 @@ phoc_output_layer_for_each_surface (PhocOutput          *self,
                                     PhocSurfaceIterator  iterator,
                                     void                *user_data)
 {
-  PhocLayerSurface *layer_surface;
+  GQueue *layer_surfaces = phoc_output_get_layer_surfaces_for_layer (self, layer);
 
-  wl_list_for_each_reverse (layer_surface, &self->layer_surfaces, link) {
-    if (layer_surface->layer != layer)
-      continue;
+  for (GList *l = layer_surfaces->tail; l; l = l->prev) {
+    PhocLayerSurface *layer_surface = PHOC_LAYER_SURFACE (l->data);
 
-    if (layer_surface->layer_surface->current.exclusive_zone <= 0)
-      phoc_output_layer_surface_for_each_surface (self, layer_surface, iterator, user_data);
-  }
-  wl_list_for_each (layer_surface, &self->layer_surfaces, link) {
-    if (layer_surface->layer != layer)
-      continue;
-
-    if (layer_surface->layer_surface->current.exclusive_zone > 0)
-      phoc_output_layer_surface_for_each_surface (self, layer_surface, iterator, user_data);
+    phoc_output_layer_surface_for_each_surface (self, layer_surface, iterator, user_data);
   }
 }
 
@@ -1231,22 +1249,32 @@ phoc_output_layer_for_each_surface (PhocOutput          *self,
  * @layer: The layer to get the surfaces for
  *
  * Get a list of [type@PhocLayerSurface]s on this output in the given
- * `layer`.
+ * `layer` in rendering order.
  *
- * Returns:(transfer container)(element-type PhocLayerSurface): The layer surfaces of that layer
+ * Returns:(transfer none): The layer surfaces of that layer
  */
-GList *
+GQueue *
 phoc_output_get_layer_surfaces_for_layer (PhocOutput *self, enum zwlr_layer_shell_v1_layer layer)
 {
-  GList *layer_surfaces = NULL;
+  PhocDesktop *desktop = phoc_server_get_desktop (phoc_server_get_default ());
   PhocLayerSurface *layer_surface;
+  PhocOutputPrivate *priv;
+  g_autoptr (GQueue) queue = NULL;
+
+  g_assert (PHOC_IS_OUTPUT (self));
+  priv = phoc_output_get_instance_private (self);
+
+  if (priv->layer_surfaces[layer])
+    return priv->layer_surfaces[layer];
+
+  queue = g_queue_new ();
 
   wl_list_for_each_reverse (layer_surface, &self->layer_surfaces, link) {
     if (layer_surface->layer != layer)
       continue;
 
     if (layer_surface->layer_surface->current.exclusive_zone > 0)
-      layer_surfaces = g_list_prepend (layer_surfaces, layer_surface);
+      g_queue_push_head (queue, layer_surface);
   }
 
   wl_list_for_each (layer_surface, &self->layer_surfaces, link) {
@@ -1254,10 +1282,88 @@ phoc_output_get_layer_surfaces_for_layer (PhocOutput *self, enum zwlr_layer_shel
       continue;
 
     if (layer_surface->layer_surface->current.exclusive_zone <= 0)
-      layer_surfaces = g_list_prepend (layer_surfaces, layer_surface);
+      g_queue_push_head (queue, layer_surface);
   }
 
-  return layer_surfaces;
+  GSList *stacks = phoc_desktop_get_layer_surface_stacks (desktop);
+  for (GSList *s = stacks; s; s = s->next) {
+    PhocStackedLayerSurface *stack = s->data;
+    PhocLayerSurface *stacked, *target;
+    GList *stacked_link, *target_link;
+
+    if (phoc_stacked_layer_surface_get_layer (stack) != layer)
+      continue;
+
+    stacked = phoc_stacked_layer_surface_get_layer_surface (stack);
+    if (!stacked)
+      continue;
+
+    if (phoc_layer_surface_get_output (stacked) != self)
+      continue;
+
+    target = phoc_stacked_layer_surface_get_target_layer_surface (stack);
+    if (!target)
+      continue;
+
+    if (phoc_layer_surface_get_output (stacked) != self)
+      continue;
+
+    if (phoc_layer_surface_get_layer (target) != phoc_layer_surface_get_layer (stacked)) {
+      g_critical ("Stacked surface and target surface not in same layer");
+      continue;
+    }
+
+    stacked_link = g_queue_find (queue, stacked);
+    g_assert (stacked_link);
+    g_queue_unlink (queue, stacked_link);
+
+    target_link = g_queue_find (queue, target);
+    g_assert (target_link);
+
+    switch (phoc_stacked_layer_surface_get_position (stack)) {
+    case PHOC_STACKED_SURFACE_STACK_BELOW:
+      g_debug ("Stacking '%s' below '%s'",
+               PHOC_LAYER_SURFACE (stacked_link->data)->layer_surface->namespace,
+               PHOC_LAYER_SURFACE (target_link->data)->layer_surface->namespace);
+      g_queue_insert_before_link (queue, target_link, stacked_link);
+      break;
+    case PHOC_STACKED_SURFACE_STACK_ABOVE:
+      g_debug ("Stacking '%s' above '%s'",
+               PHOC_LAYER_SURFACE (stacked_link->data)->layer_surface->namespace,
+               PHOC_LAYER_SURFACE (target_link->data)->layer_surface->namespace);
+      g_queue_insert_after_link (queue, target_link, stacked_link);
+      break;
+    default:
+      g_assert_not_reached ();
+    }
+  }
+
+  g_clear_pointer (&priv->layer_surfaces[layer], g_queue_free);
+  priv->layer_surfaces[layer] = g_steal_pointer (&queue);
+  return priv->layer_surfaces[layer];
+}
+
+/**
+ * phoc_output_set_layer_dirty:
+ * @self: the output
+ * @layer: The layer to marks as dirty
+ *
+ * Invalidate the ordering of layer surfaces.
+ *
+ * Moving a layer surface between layers or changing the exclusive zone
+ * might affect a layer surfaces position in the stack. Calling this function
+ * invalidates the current ordering and makes sure the ordering is recalculated
+ * on next access.
+ */
+void
+phoc_output_set_layer_dirty (PhocOutput *self, enum zwlr_layer_shell_v1_layer layer)
+{
+  PhocOutputPrivate *priv;
+
+  g_assert (PHOC_IS_OUTPUT (self));
+  priv = phoc_output_get_instance_private (self);
+
+  g_clear_pointer (&priv->layer_surfaces[layer], g_queue_free);
 }
 
 /**
@@ -1305,7 +1411,7 @@ phoc_output_drag_icons_for_each_surface (PhocOutput          *self,
  *
  * Iterate over surfaces on the output.
  */
-void
+static void
 phoc_output_for_each_surface (PhocOutput          *self,
                               PhocSurfaceIterator  iterator,
                               void                *user_data,
@@ -1957,6 +2063,10 @@ phoc_output_update_shell_reveal (PhocOutput *self)
 {
   PhocOutputPrivate *priv;
   gboolean old;
+
+  if (self == NULL)
+    return;
+
   g_assert (PHOC_IS_OUTPUT (self));
   priv = phoc_output_get_instance_private (self);
 
@@ -2105,3 +2215,12 @@ phoc_output_get_texture_filter_mode (PhocOutput *self)
 
   return WLR_SCALE_FILTER_BILINEAR;
 }
+
+
+struct wlr_output *
+phoc_output_get_wlr_output (PhocOutput *self)
+{
+  g_assert (PHOC_IS_OUTPUT (self));
+
+  return self->wlr_output;
+}
