@@ -293,10 +293,9 @@ render_view (PhocOutput *output, PhocView *view, PhocRenderContext *ctx)
 static void
 render_layer (enum zwlr_layer_shell_v1_layer layer, PhocRenderContext *ctx)
 {
-  g_autoptr (GList) layer_surfaces = NULL;
+  GQueue *layer_surfaces = phoc_output_get_layer_surfaces_for_layer (ctx->output, layer);
 
-  layer_surfaces = phoc_output_get_layer_surfaces_for_layer (ctx->output, layer);
-  for (GList *l = layer_surfaces; l; l = l->next) {
+  for (GList *l = layer_surfaces->head; l; l = l->next) {
     PhocLayerSurface *layer_surface = PHOC_LAYER_SURFACE (l->data);
 
     ctx->alpha = phoc_layer_surface_get_alpha (layer_surface);
